@@ -12,7 +12,7 @@
 ##############################################################################
 '''CGI Response Output formatter
 
-$Id: HTTPResponse.py 68780 2006-06-21 16:50:26Z tseaver $'''
+$Id: HTTPResponse.py 92518 2008-10-23 19:20:01Z tseaver $'''
 __version__ = '$Revision: 1.81 $'[11:-2]
 
 import types, os, sys, re
@@ -122,6 +122,10 @@ otherTypes = os.environ.get('DONT_GZIP_MAJOR_MIME_TYPES','').lower()
 if otherTypes:
     uncompressableMimeMajorTypes += tuple(otherTypes.split(','))
 
+_CRLF = re.compile(r'\r[\n]?')
+
+def _scrubHeader(name, value):
+    return ''.join(_CRLF.split(str(name))), ''.join(_CRLF.split(str(value)))
 
 class HTTPResponse(BaseResponse):
     """\
@@ -247,10 +251,8 @@ class HTTPResponse(BaseResponse):
         Sets an HTTP return header "name" with value "value", clearing
         the previous value set for the header, if one exists. If the
         literal flag is true, the case of the header name is preserved,
-        otherwise word-capitalization will be performed on the header
-        name on output.'''
-        name = str(name)
-        value = str(value)
+        otherwise the header name will be lowercased.'''
+        name, value = _scrubHeader(name, value)
         key = name.lower()
         if accumulate_header(key):
             self.accumulated_headers = (
@@ -259,12 +261,23 @@ class HTTPResponse(BaseResponse):
         name = literal and name or key
         self.headers[name] = value
 
+    def getHeader(self, name, literal=0):
+        '''\
+        Get a header value
+
+        Returns the value associated with a HTTP return header, or
+        "None" if no such header has been set in the response
+        yet. If the literal flag is true, the case of the header name is
+        preserved, otherwise the header name will be lowercased.'''
+        key = name.lower()
+        name = literal and name or key
+        return self.headers.get(name, None)
+
     def addHeader(self, name, value):
         '''\
         Set a new HTTP return header with the given value, while retaining
         any previously set headers with the same name.'''
-        name = str(name)
-        value = str(value)
+        name, value = _scrubHeader(name, value)
         self.accumulated_headers = (
             "%s%s: %s\n" % (self.accumulated_headers, name, value))
 
@@ -343,7 +356,7 @@ class HTTPResponse(BaseResponse):
             self.setHeader('content-type', c)
         else:
             c = self.headers['content-type']
-            if not 'charset=' in  c:
+            if c.startswith('text/') and not 'charset=' in  c:
                 c = '%s; charset=%s' % (c, default_encoding)                
                 self.setHeader('content-type', c)
 
@@ -574,8 +587,8 @@ class HTTPResponse(BaseResponse):
         Sets an HTTP return header "name" with value "value",
         appending it following a comma if there was a previous value
         set for the header. '''
-        name = str(name).lower()
-        value = str(value)
+        name, value = _scrubHeader(name, value)
+        name = name.lower()
 
         headers = self.headers
         if headers.has_key(name):
